/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.recallEstimation.UnivariateFunctionImpl;

import java.util.concurrent.Semaphore;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.commons.math4.analysis.UnivariateFunction;
import org.apache.commons.math4.analysis.integration.BaseAbstractUnivariateIntegrator;
import org.apache.commons.math4.distribution.NormalDistribution;
import sf.recallEstimation.RecallEstimator;
import sf.recallEstimation.integrating.OnePeakPositiveNormalizedFunctionIntegrator;

/**
 *
 * @author xmic
 */
public class PrRankXLessM implements UnivariateFunction {

    private final UnivariateFunction pxProbSkBitsDif;
    private final int bSkLength;
    private final int mCandidateSetSize;
    private final double[] integralValues;
    private final Double sketchesIDim;

    public PrRankXLessM(Double sketchesIDim, UnivariateFunction pxProbSkBitsDif, int bSkLength, int mCandidateSetSize, double[] mub, double[] sigmab) {
        this.pxProbSkBitsDif = pxProbSkBitsDif;
        this.bSkLength = bSkLength;
        this.mCandidateSetSize = mCandidateSetSize;
        this.sketchesIDim = sketchesIDim;
        integralValues = evaluateIntegrals(RecallEstimator.PARALELIZATION, mub, sigmab);
//        integralValues = evaluateIntegrals(mub, sigmab);
        System.err.println("b;integral_b;mu_b;sigma^2_b;sigma_b");
        System.err.println("0;" + integralValues[0] + ";" + mub[0] + ";" + (sigmab[0] * sigmab[0]) + ";" + sigmab[0]);
        for (int i = 1; i < integralValues.length; i++) {
            System.err.println(i + ";" + integralValues[i] + ";" + mub[i] + ";" + (sigmab[i] * sigmab[i]) + ";" + sigmab[i]);
            if (integralValues[i - 1] < integralValues[i]) {
                Logger.getLogger(PrRankXLessM.class.getName()).log(Level.WARNING, "Integral values increases from {0} to {1}", new Object[]{integralValues[i - 1], integralValues[i]});
            }
        }
    }

    @Override
    public double value(double x) {
        double ret = 0;
        PxbSmallBinAnalogue pxb = new PxbSmallBinAnalogue(bSkLength, sketchesIDim, pxProbSkBitsDif, x);
        for (int b = 0; b <= bSkLength; b++) {
            double pxbVal = pxb.value(b);
            ret += pxbVal * integralValues[b];
        }
        return ret;
    }

    private double[] evaluateIntegrals(double[] mub, double[] sigmab) {
        double[] ret = new double[bSkLength + 1];
        for (int i = 0; i <= bSkLength; i++) {
            NormalDistribution fYMuBSigmaB = new NormalDistribution(mub[i], sigmab[i]);
            BaseAbstractUnivariateIntegrator integrator = new OnePeakPositiveNormalizedFunctionIntegrator(mub[i]);
            ret[i] = integrator.integrate(Integer.MAX_VALUE, fYMuBSigmaB, Double.NEGATIVE_INFINITY, mCandidateSetSize);
            System.out.println("i: " + i + ", result: " + ret[i]);
        }
        Logger.getLogger(RecallEstimator.class.getName()).log(Level.INFO, "Finished evaluating integrals f(y;mub,sigmab).");
        return ret;
    }

    private double[] evaluateIntegrals(int paralelization, double[] mub, double[] sigmab) {
        double[] ret = new double[bSkLength + 1];
        Semaphore sem = new Semaphore(paralelization);
        AtomicInteger counter = new AtomicInteger(0);
        PrRankXLessMThread[] thds = new PrRankXLessMThread[bSkLength + 1];
        for (int i = 0; i <= bSkLength; i++) {
            thds[i] = new PrRankXLessMThread(counter, ret.length, mub[i], sigmab[i], mCandidateSetSize, sem);
            try {
                sem.acquire();
                new Thread(thds[i]).start();
            } catch (InterruptedException ex) {
                Logger.getLogger(RecallEstimator.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        for (int i = 0; i < paralelization; i++) {
            try {
                sem.acquire();
            } catch (InterruptedException ex) {
                Logger.getLogger(RecallEstimator.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        double max = 0;
        for (int i = bSkLength; i >= 0; i--) {
            ret[i] = Math.max(max, Math.min(1, thds[i].getResult()));
            if (ret[i] > max) {
                max = ret[i];
            }
//            System.out.println("i: " + i + ", result: " + ret[i]);
        }
        Logger.getLogger(RecallEstimator.class.getName()).log(Level.INFO, "Finished evaluating integrals f(y;mub,sigmab).");
        return ret;
    }

}
