/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.separability.measuredGeneral;

import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import main.tools.Tools;
import messif.objects.LocalAbstractObject;
import org.apache.commons.math4.analysis.BivariateFunction;
import sf.objects.impl.Sketch;
import sf.separability.QualityOfSketchingTechnique;
import sf.separabilityAnalytique.BivariateIntegratorLucia;
import sf.separabilityMeasured.QualityOfSketchingTechniqueMeasured;

/**
 *
 * @author xmic
 */
public class QualityOfSketchingTechGenlMeas extends QualityOfSketchingTechnique {

    private final BivariateFunction probOrderPreserve;
    private final Map<Integer, double[][]> conditionalDistribsForXIntervals;
//    private final float stepSize;
    private static int skippedBuckets;

    public QualityOfSketchingTechGenlMeas(Map<Integer, Double> distanceXDensity, Map<String[], Float> distsOnData, String sketchesPath, int sketchLength, String output, double maxX1, int countOfIntervals, float stepSize) {
        super(stepSize * countOfIntervals);
//        this.stepSize = stepSize;
        skippedBuckets = 0;
        conditionalDistribsForXIntervals = conditionalDistributionOfHamDist(distsOnData, sketchesPath, countOfIntervals, sketchLength, stepSize);
        Logger.getLogger(QualityOfSketchingTechGenlMeas.class.getName()).log(Level.INFO, "Skipped {0} buckets due to a low number of objects in it", skippedBuckets);
        probOrderPreserve = new ProbOrderPreserveImpl(distanceXDensity, conditionalDistribsForXIntervals, stepSize);

//        System.out.println(probOrderPreserve.value(0, 0));
//        for (int start = 0; start < 8; start++) {
//            System.out.println(probOrderPreserve.value(start * 100, 700));
//            System.out.println(probOrderPreserve.value(start * 100, 350));
//        }
    }

    public void evaluateAndPrint(String output, double maxX1, float stepSize) throws FileNotFoundException {
        double quality = evaluateNormalized(maxX1, stepSize);
        System.out.print("Normalized quality of sketching technique with max dist " + maxDistance + " on path is;");
        System.out.println(quality);
    }

    private static Map<Integer, double[][]> conditionalDistributionOfHamDist(Map<String[], Float> distsOnData, String sketchesPath, int countOfIntervals, int sketchLength, float stepSize) {
        Iterator<Map.Entry<String[], Float>> distsXIt = distsOnData.entrySet().iterator();
        SortedMap<Integer, List<Float>> distancesXAndHammings = initMapIntListFloat(countOfIntervals);
        Map<String, LocalAbstractObject> sketches = Tools.getObjectsAsLocatorMap(Sketch.class.getName(), sketchesPath, -1);
        while (distsXIt.hasNext()) {
            Map.Entry<String[], Float> next = distsXIt.next();
            float dist = next.getValue();
            if (dist >= stepSize * countOfIntervals) {
                continue;
            }
            int key = (int) (dist / stepSize);
            String[] locators = next.getKey();
            float hammDist = sketches.get(locators[0]).getDistance(sketches.get(locators[1]));
            distancesXAndHammings.get(key).add(hammDist);
        }
        Map<Integer, double[][]> ret = new HashMap<>();
        for (Iterator<Map.Entry<Integer, List<Float>>> itX = distancesXAndHammings.entrySet().iterator(); itX.hasNext();) {
            Map.Entry<Integer, List<Float>> next = itX.next();
            int idx = next.getKey();
            List<Float> hammingDistances = next.getValue();
            if (hammingDistances.size() < QualityOfSketchingTechniqueMeasured.MIN_NUMBER_OF_DISTANCES) {
                if (!hammingDistances.isEmpty()) {
                    skippedBuckets++;
                }
                hammingDistances.clear();
            }
            double[][] probBiCumBi = evaluateDiscreteDensities(hammingDistances, sketchLength);
            if (!hammingDistances.isEmpty() && Math.abs(probBiCumBi[1][sketchLength] - 1) > 1E-4) {
                throw new RuntimeException("Wrong cumulative probability for bucket with index " + idx + " (" + probBiCumBi[1][sketchLength] + ")");
            }
            ret.put(idx, probBiCumBi);
        }
        return ret;
    }

    private static Map<Integer, Double> initMapIntFloat(int countOfIntervals) {
        Map<Integer, Double> ret = new HashMap<>();
        for (int i = 0; i < countOfIntervals; i++) {
            ret.put(i, 0d);
        }
        return ret;
    }

    private static SortedMap<Integer, List<Float>> initMapIntListFloat(int countOfIntervals) {
        SortedMap<Integer, List<Float>> ret = new TreeMap<>();
        for (int i = 0; i < countOfIntervals; i++) {
            ret.put(i, new ArrayList<>());
        }
        return ret;
    }

    private static int distsCount = 0;

    public static Map<Integer, Double> evaluateDensityOfDistX(Map<String[], Float> distsOnData, int countOfIntervals, float stepSize) {
        Map<Integer, Double> ret = initMapIntFloat(countOfIntervals);
        for (Iterator<Map.Entry<String[], Float>> it = distsOnData.entrySet().iterator(); it.hasNext();) {
            Map.Entry<String[], Float> entry = it.next();
            float distX = entry.getValue();
            if (distX >= countOfIntervals * stepSize) {
                continue;
            }
            distsCount++;
            int idx = (int) (distX / stepSize);
            double value = ret.get(idx) + 1;
            ret.put(idx, value);
        }
        for (Integer key : ret.keySet()) {
            double newValue = ret.get(key) / distsCount;
            ret.put(key, newValue);
        }
//        float sum = 0;
//        System.out.println("AAAAA");
//        for (Integer key : ret.keySet()) {
//            sum += ret.get(key);
//            System.out.println(ret.get(key) + ";" + sum);
//        }
        return ret;
    }

    private static double[][] evaluateDiscreteDensities(List<Float> hammingDistances, int sketchLength) {
        double[][] ret = new double[2][sketchLength + 1];
        if (hammingDistances.isEmpty()) {
            return ret;
        }
        for (Float hamDist : hammingDistances) {
            int hamDistI = hamDist.intValue();
            ret[0][hamDistI]++;
        }
        for (int i = 0; i <= sketchLength; i++) {
            ret[0][i] = ret[0][i] / hammingDistances.size();
        }
        ret[1][0] = ret[0][0];
        for (int i = 1; i <= sketchLength; i++) {
            ret[1][i] = ret[1][i - 1] + ret[0][i];
        }
        return ret;
    }

    @Override
    protected BivariateFunction createInstanceSepDistsx1x2InHammSpace() {
        return probOrderPreserve;
    }

    @Override
    public double evaluateNormalized(double maxX1, double stepSize) {
        if (maxX1 > maxDistance) {
            throw new IllegalArgumentException("Bad limit t for integral: t = " + maxX1 + ", max distance is " + maxDistance);
        }
        double quality = 0;
        int numberOfBuckets;
        if (maxX1 <= 0) {
            numberOfBuckets = (int) (maxDistance / stepSize);
        } else {
            numberOfBuckets = (int) Math.ceil(maxX1 / stepSize);
        }
        for (int i = 1; i < numberOfBuckets; i++) {
            for (int j = 0; j < i - 1; j++) {
                quality += probOrderPreserve.value(j * stepSize, i * stepSize);
            }
        }

//        BivariateIntegratorLucia integrator;
//        if (maxX1 >= 0) {
//            integrator = new BivariateIntegratorLucia(probOrderPreserve, 0d, maxDistance, maxX1, stepSize);
//        } else {
//            integrator = new BivariateIntegratorLucia(probOrderPreserve, 0d, maxDistance, maxDistance, stepSize);
//        }
//        double quality = integrator.evaluate();
//        return quality;
        return 1 - 2 * quality;
    }

}
